/////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com   www.trbarry.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Developer      Version     Changes
//
// 2003/06/21    Tom Barry		V 0.0.9.1	Add Search Effort parm 

// 2003/06/20    Tom Barry		V 0.0.9.0	Initial test release 
/////////////////////////////////////////////////////////////////////////////

// Notes on parm values for FrameDbl(MotionThresh, Search_Effort);
//
// For fewest artfifact but also least smooth play use FrameDbl(25,1)
//
// For the most smoothness use FrameDbl(255,3) which is my own current favorite 
// and recommendation, or, if that blends too much try FrameDbl(128,3)
//
// I'm not sure FrameDbl(????,9) ever justifies the added CPU cost and it adds
// a small number of extra artifacts.



// >>>>>>>>>>>>>>>>>>>>>> REMEMBER TO FIX STMEDIANFILTER "+2" PROBLEM, if any >>>>>>>>>>>>>





#include "windows.h"
#include "avisynth.h"
#include <malloc.h>
#include <string.h>

#define BOOL unsigned int
#define BYTE unsigned char

const __int64 YMask   = 0x00ff00ff00ff00ff;
const __int64 UVMask  = 0xff00ff00ff00ff00;
const __int64 AllMask = 0xffffffffffffffff;

class FrameDbl : public GenericVideoFilter {
    int  Max_Frames;
    __int64  TThreshHoldM;
	int Search_Effort;
	int  useFrame;

    PVideoFrame src;
    PVideoFrame dst;
    PVideoFrame prevSrc;
    unsigned char* dstp;
    int src_pitch;
    int dst_pitch;
    int rowsize;
    int height;
    int FldHeight;
	bool SSEMMXenabled;
	bool DoYV12;
	BOOL GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* pPrev,
			int FldHeight);					// go H expand lines into even output lines

public:
    FrameDbl(PClip _child, int _TThreshHoldM, int _Search_Effort,
		 IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
		_int64 i;
		
		Search_Effort = _Search_Effort;
		i = _TThreshHoldM;	
        TThreshHoldM = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;

/*	    if (vi.IsYUY2())		// >>>>>>> YUY2 not supported yet
		{
			DoYV12 = false;
		}
		else */

		if (vi.IsYV12())
		{
			DoYV12 = true;
		}
	    else
		{
//			env->ThrowError("FrameDbl: Supports YUY2 and YV12 color formats only");
			env->ThrowError("FrameDbl: Supports YV12 color formats only");
		}

	    if (!SSEMMXenabled)
		{
			env->ThrowError("FrameDbl: Only SSEMMX CPU's currently supported");
		}
        vi.num_frames = vi.num_frames * 2;
        vi.SetFPS(vi.fps_numerator * 2, vi.fps_denominator);

    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);

};


PVideoFrame __stdcall FrameDbl::GetFrame(int inFrame, IScriptEnvironment* env) 
{
PVideoFrame PrevSrc;

const BYTE* srcp;
const BYTE* pPrev;
	
	useFrame = inFrame/2;
	
	// Most of our processing is the same for YUY2 or YV12.
	src = child->GetFrame(useFrame, env);
	srcp = src->GetReadPtr(PLANAR_Y);

	if (useFrame < 1 || useFrame >= Max_Frames-1 || inFrame % 2)
	{
		return src;						// don't filter first and last frames
	}
	// >>>>>>> Warning -- this code will not work for YUY2 yet !!!!

	PrevSrc = child->GetFrame(useFrame-1, env);
	pPrev = PrevSrc->GetReadPtr(PLANAR_Y);
	dst = env->NewVideoFrame(vi);
	dstp = dst->GetWritePtr(PLANAR_Y);
	src_pitch = src->GetPitch(PLANAR_Y);
	dst_pitch = dst->GetPitch(PLANAR_Y);
	rowsize = dst->GetRowSize(PLANAR_Y);
	height = dst->GetHeight();

 //env->BitBlt(dstp, dst_pitch, srcp,src_pitch, rowsize, height);
//	if (inFrame %2)
		GoFilterIt(src_pitch, dst_pitch,
			rowsize, srcp, dstp, pPrev, height);					

	// But if it is YV12 we still have to do the 2 chroma planes
	if (vi.IsYV12())
	{
		srcp = src->GetReadPtr(PLANAR_U);
		pPrev = PrevSrc->GetReadPtr(PLANAR_U);
		dstp = dst->GetWritePtr(PLANAR_U);
		src_pitch = src->GetPitch(PLANAR_U);
		dst_pitch = dst->GetPitch(PLANAR_U);
		rowsize = dst->GetRowSize(PLANAR_U);
		height = dst->GetHeight(PLANAR_U);
		GoFilterIt(src_pitch, dst_pitch,
			rowsize, srcp, dstp, pPrev, 
			height);					// go H expand lines into even output lines

		srcp = src->GetReadPtr(PLANAR_V);
		pPrev = PrevSrc->GetReadPtr(PLANAR_V);
		dstp = dst->GetWritePtr(PLANAR_V);
		src_pitch = src->GetPitch(PLANAR_V);
		dst_pitch = dst->GetPitch(PLANAR_V);
		rowsize = dst->GetRowSize(PLANAR_V);
		height = dst->GetHeight(PLANAR_V);
		GoFilterIt(src_pitch, dst_pitch,
			rowsize, srcp, dstp, pPrev, 
			height);					// go H expand lines into even output lines
	}

	__asm emms;

    return dst;
}

#define IS_SSE
#define SSE_TYPE SSE


#define BestInterp(PADDR1, PADDR2) \
__asm { \
		__asm movq	mm0, qword ptr[(PADDR1)] /* our 4 pixels */ \
		__asm movq mm1, qword ptr[(PADDR2)] /* our pixel2 value */ \
		__asm movq mm2, mm0			/* another copy of our pixel1 value */ \
		__asm pminub mm0, mm1       /* minimum value of the 2 */ \
		__asm pmaxub mm1, mm2       /* maximum value of the 2 */ \
		__asm movq mm2, mm1			/* another copy of our max value */ \
		__asm psubusb mm2, mm0 	    /* the diff */ \
		__asm pavgb mm0, mm1		/* just keep avg */ \
		__asm movq mm3, mm2			/* another copy of our our weights (diff) */ \
		__asm psubusb mm3, mm5		/* nonzero where old weights lower, else 0 */ \
		__asm pcmpeqb mm3, mm4		/* now ff where new better, else 00	*/ \
		__asm pand	mm0, mm3		/* keep only better new avg pixels */ \
		__asm pand  mm2, mm3		/* and weights */ \
		__asm pcmpeqb mm3, mm4		/* here ff where old better, else 00 */ \
		__asm pand	mm5, mm3		/* keep only better old weights */ \
		__asm pand  mm6, mm3        /* keep only better old avg pixels */ \
		__asm por	mm5, mm2		/* merge new & old weights */ \
		__asm por	mm6, mm0        /* merge new & old avg pixels */\
}




BOOL	FrameDbl::GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* _pPrev,
			int FldHeight)					// go H expand lines into even output lines
{

const BYTE* pPrev = _pPrev;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-10) >> 3);
__int64 TThreshHoldMW = TThreshHoldM;
__int64	ChromaMask	  = (DoYV12)
						? AllMask
						: YMask;

	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pit*(FldHeight-1),
		pSrc+src_pit*(FldHeight-1), rowsize);  // and last

	pDest += dst_pit;

	if (Search_Effort >= 9)
	{
		for (y=1; y <= FldHeight-2; y++)	
		{

			__asm
			{
			// Loop general reg usage
			//
			// eax - inner loop counter
			// ebx - src lines above curr 
			// ecx - src pitch
			// edx - dest
			// edi - dest
			// esi - src pixels, prev frame, 1 line up

			mov		ebx, pSrc   
			mov		esi, pPrev
			mov		ecx, src_pit 
			mov		edi, pDest 

			// do first and last qwords, just move
			movq	mm0, qword ptr[ebx+ecx]		// just move 1st 4 pixels
			movq	qword ptr[edi], mm0
			
			mov		eax, rowsize
			add		ebx, eax
			movq	mm0, qword ptr[ebx+ecx-8]		// just move last 4 pixels
			movq	qword ptr[edi+eax-8], mm0

			sub		ebx, eax						// restore
			pxor	mm4, mm4						// useful constant 000..	
			// make some adjustments to start output on 2nd qword of 2nd line
			add		ebx, 8
			add		esi, 8
			add		edi, 8
			mov	    eax, ct			// loop ct

	LoopQ_9:	
			// we will start by pretending the pixels from the curr frame were
			// an interptreted answer with a motion value specified by our threshold
			movq	mm5, TThreshHoldMW		// pretend motion was at the thrshold
			movq    mm6, qword ptr[ebx+ecx]     // with values of curr frame

	// Search for best match in 9 possible locations, where we want the least difference 
	// between curr and prev pixels, assuming one of 9 small motions
			BestInterp(ebx-1, esi+2*ecx+1)  // up left, down right
			BestInterp(ebx+1, esi+2*ecx-1)  // up right, down left
			BestInterp(ebx+2*ecx-1, esi+1)  // down left, up right
			BestInterp(ebx+2*ecx+1, esi-1)   // down right, up left   
			BestInterp(ebx+2*ecx, esi)		// down, up
			BestInterp(ebx, esi+2*ecx)		// up, down
			BestInterp(ebx+ecx-1, esi+ecx+1) // left, right
			BestInterp(ebx+ecx+1, esi+ecx-1) // right, left
			BestInterp(ebx+ecx, esi+ecx)	// center, in old and new
		
	// We now should have our best possible interpolaton in mm6
			movntq	qword ptr[edi], mm6			// and store it
			add		esi, 8
			add		ebx, 8
			add		edi, 8

			dec		eax
			jnz		LoopQ_9
			}

		// adjust for next line
		pSrc  += src_pit;
		pPrev  += src_pit;
		pDest += dst_pit;
		}
	}
	else if (Search_Effort >= 3)	
	{
		for (y=1; y <= FldHeight-2; y++)	
		{

			__asm
			{
			// Loop general reg usage
			//
			// eax - inner loop counter
			// ebx - src lines above curr 
			// ecx - src pitch
			// edx - dest
			// edi - dest
			// esi - src pixels, prev frame, 1 line up

			mov		ebx, pSrc   
			mov		esi, pPrev
			mov		ecx, src_pit 
			mov		edi, pDest 

			// do first and last qwords, just move
			movq	mm0, qword ptr[ebx+ecx]		// just move 1st 4 pixels
			movq	qword ptr[edi], mm0
			
			mov		eax, rowsize
			add		ebx, eax
			movq	mm0, qword ptr[ebx+ecx-8]		// just move last 4 pixels
			movq	qword ptr[edi+eax-8], mm0

			sub		ebx, eax						// restore
			pxor	mm4, mm4						// useful constant 000..	
			// make some adjustments to start output on 2nd qword of 2nd line
			add		ebx, 8
			add		esi, 8
			add		edi, 8
			mov	    eax, ct			// loop ct

	LoopQ_3:	
			// we will start by pretending the pixels from the curr frame were
			// an interptreted answer with a motion value specified by our threshold
			movq	mm5, TThreshHoldMW		// pretend motion was at the thrshold
			movq    mm6, qword ptr[ebx+ecx]     // with values of curr frame

	// Search for best match in 9 possible locations, where we want the least difference 
	// between curr and prev pixels, assuming one of 9 small motions
			BestInterp(ebx+ecx-1, esi+ecx+1) // left, right
			BestInterp(ebx+ecx+1, esi+ecx-1) // right, left
			BestInterp(ebx+ecx, esi+ecx)	// center, in old and new
		
	// We now should have our best possible interpolaton in mm6
			movntq	qword ptr[edi], mm6			// and store it
			add		esi, 8
			add		ebx, 8
			add		edi, 8

			dec		eax
			jnz		LoopQ_3
			}

		// adjust for next line
		pSrc  += src_pit;
		pPrev  += src_pit;
		pDest += dst_pit;
		}
	}
	else
	{
				for (y=1; y <= FldHeight-2; y++)	
		{

			__asm
			{
			// Loop general reg usage
			//
			// eax - inner loop counter
			// ebx - src lines above curr 
			// ecx - src pitch
			// edx - dest
			// edi - dest
			// esi - src pixels, prev frame, 1 line up

			mov		ebx, pSrc   
			mov		esi, pPrev
			mov		ecx, src_pit 
			mov		edi, pDest 

			// do first and last qwords, just move
			movq	mm0, qword ptr[ebx+ecx]		// just move 1st 4 pixels
			movq	qword ptr[edi], mm0
			
			mov		eax, rowsize
			add		ebx, eax
			movq	mm0, qword ptr[ebx+ecx-8]		// just move last 4 pixels
			movq	qword ptr[edi+eax-8], mm0

			sub		ebx, eax						// restore
			pxor	mm4, mm4						// useful constant 000..	
			// make some adjustments to start output on 2nd qword of 2nd line
			add		ebx, 8
			add		esi, 8
			add		edi, 8
			mov	    eax, ct			// loop ct

	LoopQ_1:	
			// we will start by pretending the pixels from the curr frame were
			// an interptreted answer with a motion value specified by our threshold
			movq	mm5, TThreshHoldMW		// pretend motion was at the thrshold
			movq    mm6, qword ptr[ebx+ecx]     // with values of curr frame

	// Search for best match in 9 possible locations, where we want the least difference 
	// between curr and prev pixels, assuming one of 9 small motions
			BestInterp(ebx+ecx, esi+ecx)	// center, in old and new
		
	// We now should have our best possible interpolaton in mm6
			movntq	qword ptr[edi], mm6			// and store it
			add		esi, 8
			add		ebx, 8
			add		edi, 8

			dec		eax
			jnz		LoopQ_1
			}

		// adjust for next line
		pSrc  += src_pit;
		pPrev  += src_pit;
		pDest += dst_pit;
		}
	}
	return 0;
}


AVSValue __cdecl Create_FrameDbl(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new FrameDbl(args[0].AsClip(), args[1].AsInt(),args[2].AsInt(), env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("FrameDbl", "cii", Create_FrameDbl, 0);
    return "`FrameDbl' MoComp Frame Doubler";
}

